/*
  Davis-Putnam

Define:
PROPAGATE	propagate the value of unit clauses before starting the search
		(useful if TRIVIALCHOICE or WEIGHTCHOICE are used)

DEBUG		print debug info
*/

#include"list.h"
#include"queue.h"
#include"mindp.h"


queue *q;



/****************************** Simplify ************************************/
/*
  Simplify: if a variable appears with one sign only in the whole set
of clauses (independently from the current assignment).
*/

int simplify( setofclauses cl, clauseindex ix, state stt, stack stk, int options ) {
  int i;
  int p;
  listofint l;

  #ifdef DEBUG
    printf("Initial simplification\n");
  #endif

  rmallqueue(q);
  
  for(i=1; i<=cl->numvar; i++) {

    if( ix->positive[i]==NULL ) 
      inqueue(q,-i);
    
    else if( ( (cl->weight[i]==0) || (stt->weighted==0) ) &&
             (ix->positive[i]!=NULL) )
      if( ix->negative[i]==NULL )
        inqueue(q,i);
  }

  if( isempty(q) )
    return 0;

  #ifdef DEBUG
    printqueue(q);
  #endif
  
  if( unitpropagation(q, cl, ix, stt, stk, options )==1 )
       return 1;

  if( options & SELFTEST ) {
    selftest(cl,ix,stt,stk,options);
    consselftest(cl,ix,stt,stk,options);
  }

  #ifdef DEBUG
    printf("End initial simplification\n");
  #endif

  return 0;
}

/**************************** End Simplify **********************************/


/********************************** Sat *************************************/
/* 
  Returns one model of a set of clauses.
*/

int dpnext( setofclauses cl, clauseindex ix, state stt, stack stk, int options ) {
  int u;
  int lit;
  int ret;

  while(1) {
    #ifdef DEBUG
      printstate( stt );
      fflush(NULL);
    #endif

    rmallqueue(q);
    inqueue(q,stk->elem[ stk->stackindex ].literal);
    u=unitpropagation(q, cl, ix, stt, stk, options);

    if( options & SELFTEST )
      selftest( cl, ix, stt, stk, options );

    #ifdef DEBUG
      printf("\nUnit propagation result = %d\n",u);
      printstate( stt );
      printstack( stk );
      fflush(NULL);
    #endif

    if( u==1 ) {
      ret=1;
      break;
    }
    else if( u==-1 ) {
      if( backtracking( cl, ix, stt, stk, options ) == -1 ) {
        ret=-1;
        break;
      }
    }
    else {
      stt->branch++;
      if( options & SELFTEST )
        consselftest( cl, ix, stt, stk, options );
      lit=choice( cl, ix, stt, stk, options );
      addlevel( stk, lit, 0, options );
    }
  }

  return ret;
}


/*
  First model of a formula
*/

int firstmodel( setofclauses cl, clauseindex *ix, state *stt, stack *stk, int options ) {
  int u;
  int newoptions;

  newoptions=options&~SIMPLIFY;

  initial( cl, ix, stt, stk, newoptions );

  if( newoptions & PROPAGATE ) {
    u=propagateunitclauses( cl, *ix, *stt, *stk, newoptions );
    if( u!=0 )
      return u;
  }

  firstlevel( cl, *ix, *stt, *stk, newoptions );

  return dpnext( cl, *ix, *stt, *stk, newoptions );
}


/*
  Next model of a formula
*/

int nextmodel( setofclauses cl, clauseindex ix, state stt, stack stk, int options ) {
  int newoptions;

  newoptions=options&~SIMPLIFY;

  if( backtracking( cl, ix, stt, stk, newoptions ) == -1 )
    return -1;

  return dpnext( cl, ix, stt, stk, newoptions );
}


/*
  Check satisfiability and return one model
*/

int onemodel( setofclauses cl, state *stt, int options ) {
  clauseindex ix;
  stack stk;
  int u;

  initial( cl, &ix, stt, &stk, options );

  if( options | PROPAGATE ) {
    u=propagateunitclauses( cl, ix, *stt, stk, options );
    if( u!=0 )
      return u;
  }

  if( options | SIMPLIFY ) {
    if( simplify(cl,ix,*stt,stk,options)==1 )
      return 1;
  }

  firstlevel( cl, ix, *stt, stk, options );

  u=dpnext( cl, ix, *stt, stk, options );
  
  freeindex(ix);
  freestack(stk);

  return u;
}


/*
  Just verify whether the formula is satisfiable. I do not care about the
model, if any.
*/

int sat( setofclauses cl, int options ) {
  state stt;
  int u;
  
  u=onemodel( cl, &stt, options );

  freestate( stt );

  return u;
}

/******************************** End Sat ***********************************/


