/*
  This example involves two separate procedure for testing the satisfiability
of a set of clauses. The first one is the regular Davis-Putnam procedure as
implmeneted in libdp.a (function sat()). The second one is the random test
of models: we generate MAXMODELS random models, testing the satisfiability of
each.

  This program terminates when either procedure terminates. There are four
possible scenarios:

1. dp finds a model
2. dp proves unsatisfiability
3. random procedure finds a model
4. random procedure reaches the maximum count

In case 1 and 3 the formula is proved satisfiabile. In case 2 it is proved
unsatisfiable, and in case 4 nothing is known for sure.

The two procedures are implemented as two separates processes, child()
(which runs the DP procedure) and parent() (which runs the random search).

The parent set a model at random, and verify whether it is a model. If it
is, the parent sends a signal to the child, prints that the formula is
satisfiable, and ends. At each step, the shared memory is checked: if the
child is finished, the parents also ends.

The child simplty starts the DP procedure. If it a signal is received
before DP is completed, it simply dies. If DP is finished, the result is
written in the shared memory (this way, the parent process is notified
that the child is finished).

Apparently, the DP procedure almost always terminates before the random
procedure. However, if the number of variables is very small and the
formula is underconstrained, DP is slowed down by the need of allocating
memory, while the parent finds a model very quickly. Try for example 5
variables and 10 clauses.
*/

#include<stdio.h>
#include<unistd.h>
#include<sys/types.h>
#include<sys/shm.h>
#include<signal.h>

#include"dp.h"

				/* maximum number of random models to try */

#define MAXMODELS 10000


				/* set of clauses */

setofclauses cl;


				/* model (in libdp.a models are represented in a different way) */

int model[MAXVAR+1];


				/* shared memory */

int sid;
int *s;


				/* remove shared memory */

void rmshared() {
  struct shmid_ds dummy_shmctl;

  fprintf(stderr,"Removing shared memory\n");

  shmctl(sid,IPC_RMID,&dummy_shmctl);
}


				/* signal handler */

void terminate( int x ) {
  rmshared();

  exit(1);
}


				/* child signal handler */

void childterminate( int x ) {
  exit(1);
}


/* The child:
it starts dp, and when ends write the result in the shared memory.
If the parent terminates before the search is over, it sends a USR1
signal to the child, which terminates.
*/ 

int child() {
  int x;

  x=sat(cl,DEFAULTDP);

  *s=x;

  return 0;
}


/* The parent:
test models at random. If it finds a model of the set of clauses, it
sends a SIGUSR1 to the child (which terminates). Each time it polls
the shared memory to see whether the child has terminated or not (when
the child is finished, it sets *s to something different than 0).
If the child has terminated, the parent prints the result (which is
written in the shared memory *s), and exit immediately.

The parameter is the id of the child.
*/

int parent( int id ) {
  int t,i,c;
  int found,allsat,sat;
  listofint l;

  found=0;

  for(t=0; t<MAXMODELS; t++) {
    /* printf("Try n. %d\n",t); */

					/* random model */
    for(i=1; i<=cl->numvar; i++)
      model[i]=rand()%2;


					/* check whether all the clauses are satisfied */
    allsat=1;

    for(c=0; c<cl->numclauses; c++) {

      sat=0;

      for(l=cl->cl[c].lett; l!=NULL; l=l->next)
        if( ( (l->elem>0) && (model[abs(l->elem)]==1) ) ||
	    ( (l->elem<0) && (model[abs(l->elem)]==0) ) ) {
          sat=1;
          break;
        }

      if( sat==0 ) {
        allsat=0;
        break;
      }
    }
  
					/* if they are, stop */
    if( allsat==1 ) {
      found=1;
      break;
    }

					/* check the child */
    if( *s!=0 )
      break;
  }


					/* terminates the child
					  id=-1 when we test the parent alone */
  if( id>-1 )
    kill( id, SIGUSR1 );

		
					/* print answer */
  if( found )
    printf("Satisfiable (random procedure)\n");
  else if( *s==1 )
    printf("Satisfiable (dp)\n");
  else if( *s==-1)
    printf("Unsatisfiable (dp)\n");
  else
    printf("Maximum count reached\n");


                        /* remove shared memory */
  rmshared();

  return 0;
}




main() {
  int id;
  struct shmid_ds dummy_shmctl;


				/* read clauses */

  cl=newsetofclauses();
  readclauses(stdin,&cl);


  #ifdef PARENTALONE
				/* run parent alone (this is only for testing the parent) */
  s=(int *) malloc( sizeof(int) );
  *s=0;
  parent(-1);

  return 0;
  #endif



				/* set termination signals */

  signal( SIGINT, terminate );
  signal( SIGTERM, terminate );


                        	/* allocate shared memory */
  				/* this is "equivalent" to s=(int *) malloc( sizeof(int) ); */

  sid=shmget(IPC_PRIVATE,sizeof(int), IPC_CREAT | SHM_R | SHM_W);       /* allocate a shared memory */
  s=(int *) shmat(sid,0,0);                                             /* s is a pointer to that block */


                        	/* set initial value (no model is found) */

  *s=0;


				/* I put this here because the parent may terminate
				   before the child is started... */

  signal( SIGUSR1, childterminate );


                        	/* fork */

  id=fork();
  switch( id ) {
    case 0:     child();
                break;
    default:    parent( id );
                break;
  }

  return 0;
}

