/*
  The satisfiability socket. Opens a socket, and bind it to the port 9090.
It only accept one connection at time. When a connection starts, a set of
clauses is read, its satisfiability is checked, and the result is printed
back to channel.

The satisfiability check is done by the child. The parent wait unil either
the child is finished, or MAXSEC seconds are elapsed. Either way, the
result is printed back to the channel.

Start the program, then type
	telnet localhost 9090

Do not kill this program using kill -9 or kill -KILL, as this prevent the
deallocation of the shared memory. Use kill -TERM (or just kill pid), or
control-C instead.
*/

#include<stdio.h>
#include<unistd.h>
#include<sys/types.h>
#include<sys/shm.h>
#include<signal.h>
#include<sys/socket.h>
#include<netinet/in.h>
#include<netdb.h>
#include<fcntl.h>


#include"dp.h"

			/* max number of seconds each search */
#define MAXSEC 10

			/* port */

#define PORT 9090

			/* shared memory */
int *s;
int sid;


			/* communication channel */
int c;
FILE *in,*out;


/*
  Things to do before terminating:
remove shared memory.
*/

int onexit() {
  struct shmid_ds dummy_shmctl;

  fprintf(stderr,"Removing shared memory\n");

  shmctl(sid,IPC_RMID,&dummy_shmctl);
}


/*
  Print the result of the satisfiability procedure
*/

void printresult() {
  if( s[0]==1 )
    fprintf(out,"xxx Satisfiable\n");
  else if( s[0]==-1 )
    fprintf(out,"xxx Unsatisfiable\n");
  else
    fprintf(out,"xxx Search stopped due to time limits\n");
  fflush(out);
}


/*
  Signal handlers
*/

void terminate( int s ) {
  onexit();
  exit(1);
}

void childterminate( int s ) {
  printresult();
  close(c);
  exit(1);
}


/*
  Check satisfiability
*/

int satisfiable( setofclauses cl ) {
  int u;

  u=sat(cl,DEFAULTDP);

  s[0]=u;

  return u;
}


/*
  Main
*/

main() {
  int sock;				/* socket */
  struct sockaddr_in sa_server;		/* address */
  int one=1;

  int id;				/* pid of the child */
  setofclauses cl;			/* set of clauses */
  int l;				/* last literal read */
  char buf[5000];
  int i;


                                /* set termination signals */

  signal( SIGINT, terminate );
  signal( SIGTERM, terminate );


                                /* allocate shared memory */
                                /* this is "equivalent" to s=(int *) malloc( 2 * sizeof(int) ); */

  sid=shmget(IPC_PRIVATE, 2*sizeof(int), IPC_CREAT | SHM_R | SHM_W);       /* allocate a shared memory */
  s=(int *) shmat(sid,0,0);                                             /* s is a pointer to that block */


                                /* I put this here for no reason */

  signal( SIGUSR1, childterminate );


				/* open socket */

  sock=socket( AF_INET, SOCK_STREAM, 0 );
  if( sock<0 ) {
    perror("Error in opening socket");
    exit(1);
  }
  if ( setsockopt( sock, SOL_SOCKET, SO_REUSEADDR, (char *)&one, sizeof(one) ) )
    perror("Set option value");


				/* set address and bind */

  sa_server.sin_family=AF_INET;
  sa_server.sin_addr.s_addr=htonl(INADDR_ANY);
  sa_server.sin_port=htons(PORT);
  
  if( bind( sock, (struct sockaddr *) &sa_server, sizeof(sa_server) ) == -1 ) {
    perror("Error in bind");
    exit(1);
  }


				/* listen */

  if( listen( sock, 5 ) == -1 ) {
    perror("Error in listen");
    exit(1);
  }


				/* main loop: do not use kill -KILL or kill -9 to terminate */

  while( 1 ) {

				/* accept */
    c=accept( sock, 0, 0 );
    if( c<0 ) {
      perror("Accept");
      exit(1);
    }

				/* initialize connection */

    in=fdopen(c, "r");
    out=fdopen(c, "w");

    fprintf(out,"xxx-Welcome to the satisfiability socket\n");
    fprintf(out,"xxx-Insert a set of clauses. Terminate each clause with 0.\n");
    fprintf(out,"xxx Terminate the set with %%\n");
    fflush(out);
  

				/* input the set of clauses */
    cl=newsetofclauses();
    readclauses( in, &cl );

    fprintf(out,"xxx-Running satisfiability procedure.\n");
    fflush(out);

    printclauses(cl);


				/* fork */

    s[0]=0;

    id=fork();
    if( id==0 ) {
      satisfiable(cl);
      printresult();
      close(c);
      return 0;
    }


				/* parent: sleep and kill the child if is not finished */

    close(c);

    for(i=0; i<MAXSEC; i++) {
      sleep(1);
      if( s[0]!=0 )
        break;
    }

    if( s[0]==0 )
      kill( id, SIGUSR1 );

    freeclauses(cl);
  }

  onexit();

  return 0;
}

