/*
  Yet another example with two concurrent processes. In this example,
we first determine the literal appearing most often in the set of
clauses. Then, we fork two children. The first child tries to find
a model for theory u {m}, while the second one tries theory u {-m},
where m is the literal appearing most often in the theory.

After forked the children, the parent goes to sleep for 1 second.
Then, it checks whether any of the two children has finished. This
information is in the shared memory: the first child writes the
answer in s[0] when terminates; the second one in s[1].

The parent repeats this cycle unless: a. a child has found a model;
or b. both children have proved unsatisfiabiilty; or c. the maximum
time count is reached (that is, we have already tried for MAXSEC seconds).

The parent then terminates the children that are already running (if
any) and writes the answer. In case a. or b. the answer is satisfiability
and unsatisfiability, respectively. In case c. no conclusion can be
drawn (we just write whether a child has terminated or not).
*/

#include<stdio.h>
#include<unistd.h>
#include<sys/types.h>
#include<sys/shm.h>
#include<signal.h>

#include"dp.h"

				/* max number of seconds */
#define MAXSEC 10

				/* shared memory */
int *s;
int sid;


/*
  Determine the literal appearing most often in the theory.
*/

#define COMBINE(x,y)	((x)*(y)*1024+(x)+(y))

int mostoften( setofclauses cl ) {
  int positive[MAXVAR+1];
  int negative[MAXVAR+1];
  int c,i;
  listofint l;
  int bestval,bestvar;

  memset( positive, 0, (cl->numvar+1)*sizeof(int) );
  memset( negative, 0, (cl->numvar+1)*sizeof(int) );


				/* determine positive and negative occurrences */

  for(c=0; c<cl->numclauses; c++)
    for(l=cl->cl[c].lett; l!=NULL; l=l->next)
      if( l->elem > 0 )
        positive[l->elem]++;
      else
        negative[-l->elem]++;


				/* find the maximum literal according to COMBINE */

  bestval=-1;
  for(i=1; i<=cl->numvar; i++)
    if( COMBINE(positive[i],negative[i]) > bestval ) {
      bestval=COMBINE(positive[i],negative[i]);
      bestvar=i;
    }

  return bestvar;
}


/*
  Things to do before terminating:
remove shared memory.
*/

int onexit() {
  struct shmid_ds dummy_shmctl;

  fprintf(stderr,"Removing shared memory\n");

  shmctl(sid,IPC_RMID,&dummy_shmctl);
}


/*
  Signal handlers
*/

void terminate( int s ) {
  onexit();
  exit(1);
}

void childterminate( int s ) {
  exit(1);
}


/*
  Check satisfiability using literal first.
*/

int trysat( setofclauses cl, int first, int result ) {
  int u;

  printf("Determining satisfiability using literal %d\n",first);

  addliteral( first, &cl );
  endclause( &cl );

  u=sat(cl,DEFAULTDP);

  printf("Theory+{%d}: ",first);
  if( u==1 )
    printf("satisfiable\n");
  else
    printf("unsatisfiable\n");

  s[result]=u;

  return 0;
}


/*
  Main
*/

int main() {
  int m;		/* most often literal */
  int c1,c2;		/* pid of the children */
  setofclauses cl;	/* set of clauses */
  int i;

                                /* set termination signals */

  signal( SIGINT, terminate );
  signal( SIGTERM, terminate );


                                /* allocate shared memory */
                                /* this is "equivalent" to s=(int *) malloc( 2 * sizeof(int) ); */

  sid=shmget(IPC_PRIVATE, 2*sizeof(int), IPC_CREAT | SHM_R | SHM_W);       /* allocate a shared memory */
  s=(int *) shmat(sid,0,0);                                             /* s is a pointer to that block */


                                /* set initial value (no model is found) */

  s[0]=0;
  s[1]=0;


                                /* I put this here for no reason */

  signal( SIGUSR1, childterminate );


				/* read clauses */

  cl=newsetofclauses();
  readclauses( stdin, &cl );


				/* determine the literal appearing most often */

  m=mostoften( cl );
  printf("Variable appearing most often in the theory: %d\n",m);


				/* fork the children */

  c1=fork();

  if( c1==0 ) {
    trysat( cl, m, 0 );
    return 0;
  }

  c2=fork();

  if( c2==0 ) {
    trysat( cl, -m, 1 );
    return 0;
  }


				/* each second, check whether children are finished */

  for(i=0; i<MAXSEC; i++) {
    sleep(1);

				/* both have concluded */
    if( (s[0]!=0) && (s[1]!=0) )
      break;

				/* child 1 has found a model */
    if( s[0]==1 )
      break;

				/* child 2 has found a model */
    if( s[1]==1 ) 
      break;
  }


				/* kill the surviving children (if any) */

  if( s[0]==0 ) 
    kill( c1, SIGUSR1 );

  if( s[1]==0 )
    kill( c2, SIGUSR1 );



				/* print the result */

  if( (s[0]==1) || (s[1]==1) )
    printf("Satisfiable\n");
  else if ( (s[0]==-1) && (s[1]==-1) )
    printf("Unsatisfiable\n");
  else {
    printf("Search terminated before proof is reached.\n");
    if( s[0]==-1 )
      printf("Child 1 proved inconsistency\n");
    if( s[1]==-1 )
      printf("Child 2 proved inconsistency\n");
  }

  onexit();

  return 0;
}




