/*
  Initialization functions.

Define:
PROPAGATE	evaluate and propagate unary clauses.
*/

#include"list.h"
#include"queue.h"
#include"mindp.h"

queue *q;


/***************************** Create Index *******************************/
/*
Creates the index of the clauses ix.
The index ix is a stucture with two elements:

ix->positive
ix->negative

Each is an array of lists of int, meaning:

ix->positive[v] is the list of the clauses in which v appears positive
ix->negative[v] is the list of the clauses in which v is negated.

The variables numpositive and numnegative indicates how many clauses
the literal is contained in.
*/

clauseindex createindex( setofclauses cl ) {
  int c;
  listofint l;
  int v;
  clauseindex ix;


			/* allocate the index */

  ix=(clauseindex) malloc( sizeof(*ix) );
  if( ix==NULL ) {
    fprintf(stderr,"Malloc failed in createindex()\n");
    exit(1);
  }

  ix->numvar=cl->numvar;

			/* all the lists are empty */

  for(v=1; v<=cl->numvar; v++) {
    ix->numnegative[v]=0;
    ix->negative[v]=NULL;
    ix->numpositive[v]=0;
    ix->positive[v]=NULL;
  }


			/* add each clause to the lists of its literals */

  for(c=0; c<cl->numclauses; c++)
    for(l=cl->cl[c].lett; l!=NULL; l=l->next) {
      v=l->elem;
      if( v<0 ) {
        ix->numnegative[-v]++;
        ix->negative[-v]=conslistofint(ix->negative[-v],c);
      }
      else {
        ix->numpositive[v]++;
        ix->positive[v]=conslistofint(ix->positive[v],c);
      }
    }

  return ix;
}

/*************************** End Create Index *****************************/


/****************************** Init State ********************************/
/*
  Create the initial state.
*/

state initstate( setofclauses cl, clauseindex ix ) {
  state s;
  listofint l;
  int x;
  
							/* allocate memory */

  s=(state) malloc( sizeof(*s) );
  if( s==NULL ) {
    fprintf(stderr,"Malloc failed in initstate()\n");
    exit(1);
  }

							/* numvar and numclauses */
  s->numvar=cl->numvar;
  s->numclauses=cl->numclauses;

							/* assignment */
  memset(s->val, 0, cl->numvar * sizeof(char) );

							/* true and false count */
  memset(s->truelits, 0, cl->numclauses * sizeof(int) );
  memset(s->falselits, 0, cl->numclauses * sizeof(int) );

							/* by default, no weight */
  s->posclause=NULL;
  s->weighted=0;

						/* binary and occurrence count */
  s->binary=0;

  for(x=1; x<=cl->numvar; x++) {
    s->pos[x]=ix->numpositive[x];

    s->binarypos[x]=0;
    for(l=ix->positive[x]; l!=NULL; l=l->next)
      if( cl->cl[l->elem].numlett==2 ) {
        s->binarypos[x]++;
        s->binary++;
      }

    s->neg[x]=ix->numnegative[x];

    s->binaryneg[x]=0;
    for(l=ix->negative[x]; l!=NULL; l=l->next)
      if( cl->cl[l->elem].numlett==2 ) {
        s->binaryneg[x]++;
        s->binary++;
      }
  }

  s->binary=s->binary/2;

  s->found=0;

  s->branch=0;

  return s;
}

/**************************** End Init State ******************************/


/****************************** Init Stack ********************************/
/*
  Create the stack, with the first level set
*/

stack initstack() {
  stack stk;

  stk=(stack) malloc( sizeof(*stk) );
  if( stk==NULL ) {
    fprintf(stderr,"Malloc failed in initstack()\n");
    exit(1);
  }

  stk->elem[0].firstdeduced=0;
  stk->elem[0].numdeduced=0;

  stk->elem[0].weight=0;

  stk->elem[0].firstunassweighted=1;

  stk->elem[0].numtrueclauses=0;

  stk->stackindex=0;

  return stk;
}

/**************************** End Init Stack ******************************/


/********************************* Init ***********************************/
/*
1. create the index of the clauses
2. create the initial state
3. create the stack
*/

int initial( setofclauses cl, clauseindex *ix, state *stt, stack *stk, int options ) {
  int c;
  int u;

  q=emptyqueue();

  *ix=createindex( cl );

  *stt=initstate( cl, *ix );

  *stk=initstack( );

  #ifdef DEBUG
    printf("Cheking initial structures...");
  #endif

  selftest( cl, *ix, *stt, *stk, options & ~POSITIVE );
  consselftest( cl, *ix, *stt, *stk, options & ~POSITIVE );

  #ifdef DEBUG
    printf("ok\n");
  #endif

  return 0;
}


/*
  Propagate unit clauses
*/

int propagateunitclauses( setofclauses cl, clauseindex ix, state stt, stack stk, int options ) {
  int c,u;

  rmallqueue(q);

  for(c=0; c<cl->numclauses; c++)
    if( cl->cl[c].numlett==1 )
      inqueue(q,cl->cl[c].lett->elem);

  #ifdef DEBUG
    printf("Unit clauses to propagate: ");
    printqueue(q);
  #endif

  if( !isempty(q) ) {
    u=unitpropagation(q, cl, ix, stt, stk, options);
    if( u!=0 )
      return u;
  }

  #ifdef DEBUG
    printf("Model: ");
    printstate( stt );
    fflush(NULL);
  #endif

  return 0;
}


/*
  Create the first level of the stack
*/

int firstlevel( setofclauses cl, clauseindex ix, state stt, stack stk, int options ) {
  int c;
  
  c=choice( cl, ix, stt, stk, options );

  addlevel( stk, c, 0, options );

  stt->branch++;	/* the call to choice above is a branch */

  return 0;
}

/******************************* End Init *********************************/



