/*
  Minimal models using Davis-Putnam
Heuristics:
1. fail whenever the weight of a model is greater or equal than the last best found
2. use positive clauses to infer a lower bound of the weight
2. all the variables whose weight is more than the current difference are set to false

Note: it must hold weight*1000 < overflow.

Define:

USESIGNALS      compile with the functions for handling signals

*/

#include"dp.h"

#include<signal.h>
#include<sys/times.h>

#define PIDFILENAME "pid.mindp"



/******************************** Minimal ***********************************/
/*
  Functions for determining minimal models using logarithmic search.
*/

/*
  Find the first minimal model
*/

int findfirstlogweight( setofclauses cl, clauseindex *ix, state *sttlast, stack *stklast, int options ) {
  state stt;		/* state at the last model found */
  stack stk;		/* stack at the last model found */
  int k,kmax,kmin,kinter,klast;
  int x;
  int u;


				/* allocate memory for the last state and stack */

  *sttlast=(state) malloc( sizeof(**sttlast) );
  if( *sttlast==NULL ) {
    fprintf(stderr,"Malloc failed in mindp() while allocating sttlast\n");
    exit(1);
  }
  *stklast=(stack) malloc( sizeof(**stklast) );
  if( *stklast==NULL ) {
    fprintf(stderr,"Malloc failed in mindp() while allocating stklast\n");
    exit(1);
  }


				/* initial bounds (first() will check if something better
				   can be done) */

  kmax=0;
  for(x=1; x<=cl->numvar; x++)
    kmax+=cl->weight[x];
  kmin=0;
  

				/* find first model (if any) */

  #ifdef DEBUG
    printf("\n***********************\n");
    printf("* FINDING FIRST MODEL *\n");
    printf("***********************\n\n");
    printf("KMAX=%d KMIN=%d\n\n",kmax,kmin);
  #endif

  (*sttlast)->intertime=NULL;
  (*sttlast)->intertime=conslistofint((*sttlast)->intertime,usertime());
  (*sttlast)->intermod=NULL;

  k=first( cl, ix, &stt, &stk, kmax, options );

  stt->found=1;
  stt->intertime=conslistofint((*sttlast)->intertime,usertime());
  stt->intermod=conslistofint((*sttlast)->intermod,k);

  copystate( *sttlast, stt );
  copystack( *stklast, stk );

  if( k<0 )
    return k;

  #ifdef INTER
    printf("K=%d (BOUND=%d) TIME=%d\n",k,kmax,(*sttlast)->intertime->elem);
    /* printweighted( cl, stt ); */
  #endif

  klast=k;
  kmax=k-1;
  kinter=(kmax+kmin)/2;


				/* main loop */

  while( kmax>=kmin ) {
    #ifdef DEBUG
      printf("\n**********************\n");
      printf("* FINDING NEXT MODEL *\n");
      printf("**********************\n\n");
      printf("KMAX=%d KMIN=%d KINTER=%d\n\n",kmax,kmin,kinter);
    #endif


                                /* next model */
    k=next( cl, *ix, stt, stk, kinter, options );

    #ifdef INTER
      printf("K=%d (BOUND=%d) ",k,kinter);
    #endif

    if( k<0 ) {
      kmin=kinter+1;
      (*sttlast)->intertime=conslistofint((*sttlast)->intertime,usertime());
      (*sttlast)->intermod=conslistofint((*sttlast)->intermod,k);

      copystate( stt, *sttlast );
      copystack( stk, *stklast );
    }
    else {
      kmax=k-1;
      stt->kmax=k-1;
      klast=k;
      stt->intertime=conslistofint((*sttlast)->intertime,usertime());
      stt->intermod=conslistofint((*sttlast)->intermod,k);

      copystate( *sttlast, stt );
      copystack( *stklast, stk );
    }

    kinter=(kmax+kmin)/2;
    

    #ifdef INTER
      printf("TIME=%d\n",(*sttlast)->intertime->elem);
      /* printweighted( cl, stt ); */
    #endif
  }

  return klast;
}




/*
  Only the first model is necessary
*/

int logweight( setofclauses cl, state *sttlast, int options ) {
  clauseindex ix;
  stack stklast;
  int u;

  u=findfirstlogweight( cl, &ix, sttlast, &stklast, options );

  freeindex(ix);
  freestack(stklast);

  return u;
}


/*
  Find first minimal model
*/

int firstlogweight( setofclauses cl, clauseindex *ix, state *sttlast, stack *stklast, int options ) {
  return findfirstlogweight( cl, ix, sttlast, stklast, options&~SIMPLIFY );
}


/*
  Next model with minimal weight
*/

int nextlogweight( int kmax, setofclauses cl, clauseindex ix, state stt, stack stk, int options ) {
  return next( cl, ix, stt, stk, kmax, options&~SIMPLIFY );
}

/****************************** End Minimal *********************************/



                                /* signal handlers */

void onexit() {
  unlink(PIDFILENAME);
}

#ifdef USESIGNALS

  void usersignal( int s ) {
    printstatus();

    signal( SIGUSR1, usersignal);
  }

  void exitsignal( int s ) {
    printf("\nSignal %d received. Exiting.\n",s);
    // printstatus(s);
    onexit();

    exit(1);
  }
#endif


main( int argc, char *argv[] ) {
  int c;
  setofclauses cl;
  clauseindex ix;
  state stt;
  stack stk;
  int k;
  listofint l;

  FILE *pidfile;
  int starttime, leaptime, endtime;


  cl=newsetofclauses();
  readclauses( stdin, &cl );

  /* printclauses( cl ); */

  #ifdef DEBUG
    printclauses( cl );
  #endif

  #ifdef USESIGNALS
    pidfile=fopen(PIDFILENAME,"w");
    fprintf(pidfile,"%d",getpid());
    fclose(pidfile);
  
    signal( SIGTERM, exitsignal );
    signal( SIGINT, exitsignal );
    signal( SIGUSR1, usersignal );
  #endif

  if( argc>1 ) {
    if( onemodel(cl,&stt,DEFAULTDP)==1 )
      printf("SAT\n");
    else
      printf("UNSAT\n");

    printf("%d\n",stt->branch);

    return 0;
  }

  starttime=usertime();
  k=logweight( cl, &stt, DEFAULTDP );
  endtime=usertime();

  if( k<0 )
    printf("Inconsistente\n");
  else
    printf("Consistente: peso=%d\n",k);

//  printf("%d\n",stt->branch);

//  fprintf(stderr,"nodes: %d\n",stt->branch);

  fprintf(stderr,"%4d ",endtime-starttime);

#ifdef NO
  if( stt->intertime->next==NULL ) 
    fprintf(stderr,"%4d\n",endtime-starttime);
  else {
    fprintf(stderr,"%4d ",endtime - stt->intertime->elem );
    for(l=stt->intertime; l->next!=NULL; l=l->next)
      fprintf(stderr,"%4d ",l->elem - l->next->elem );
    fprintf(stderr,"\n");
  }

  fprintf(stderr,"%4d %4d ",k,k);
  for(l=stt->intermod; l!=NULL; l=l->next)
    fprintf(stderr,"%4d ",l->elem);
#endif

  fprintf(stderr,"\n");

  onexit();

  return 0;
}

