/*
  Function for evaluating a lower bound of the weight.

POSITIVE	use heuristics on positive clauses

DEBUG		print debug info
*/


#include"list.h"
#include"queue.h"
#include"mindp.h"



/*************************** Weight Lower Bound *****************************/
/*
  Evaluates a lower bound of the weight.
sign=-1		evaluates weight of negative literals
sign=1		weight of positive literals

If sign=1 we have the list of candidates in ix. Otherwise, all the clauses
must be checked (sign=-1 is currently used only for the initial evaluation,
of k, so there is no need of building an index for the negative clauses.)
*/


/*
  Lower bound determined over the set of all clauses
*/

int allclauseslowerbound( setofclauses cl, clauseindex ix, state stt, int sign ) {
  int numpositive;
  int numcand;
  int positive[MAXCLAUSES];     /* positive clauses */
  int c,d,pos,x;
  listofint l;
  int best;
  int w;
  int quanti[MAXVAR+1];
  int i;

  if( abs(sign)!=1 ) {
    printf("Sign in weightolowerbound() must be either -1 or +1\n");
    exit(1);
  }



  memset( quanti, 0, (cl->numvar+1) * sizeof(int) );

  numpositive=0;

  				/* build the set if positive clauses and the array
				   of occurencies of variables */

  for(c=0; c<cl->numclauses; c++) {
    d=c;

    if( stt->truelits[d]==0 ) {

      pos=1;
                                                /* check for negative literals */
      for(l=cl->cl[d].lett; l!=NULL; l=l->next)
        if( ( (sgn(l->elem)==-sign) || (cl->weight[abs(l->elem)]==0) ) &&
            (stt->val[abs(l->elem)]==0) ) {
          pos=0;
          break;
        }

                                                /* add to the count of clauses and literals */
      if( pos ) {
        for(l=cl->cl[d].lett; l!=NULL; l=l->next)
          if( (sgn(l->elem)==sign) && (stt->val[abs(l->elem)]==0) )
            quanti[abs(l->elem)]++;
        positive[numpositive]=d;
        numpositive++;
      }

    }
  }


                                /* for each clause, take the literal with the best fraction
                                   weight/number of occurrencies */

  w=0;

  for(c=0; c<numpositive; c++) {
    d=positive[c];

    best=1000000;

    for(l=cl->cl[d].lett; l!=NULL; l=l->next)
      if( stt->val[abs(l->elem)]==0 )
        if( cl->weight[abs(l->elem)] * 1000 / quanti[abs(l->elem)] < best )
           best= cl->weight[abs(l->elem)] * 1000 / quanti[abs(l->elem)];

    w+=best;
  }

  return (w+990)/1000;
}


/*
  Lower bound based on the set of candidate clauses
*/

int candidatelowerbound( setofclauses cl, clauseindex ix, state stt, int sign ) {
  int numpositive;
  int numcand;
  int positive[MAXCLAUSES];	/* positive clauses */
  int c,d,pos,x;
  listofint l;
  int best;
  int w;
  int quanti[MAXVAR+1];


  if( abs(sign)!=1 ) {
    printf("Sign in weightolowerbound() must be either -1 or +1\n");
    exit(1);
  }



  memset( quanti, 0, (cl->numvar+1) * sizeof(int) );

  numpositive=0;

				/* build the set if positive clauses and the array
				   of occurencies of variables */

  for(c=0; c<ix->numcand; c++) {

    d=ix->cand[c];

    if( stt->truelits[d]==0 ) {

      pos=1;
						/* check for negative literals */
      for(l=cl->cl[d].lett; l!=NULL; l=l->next)
        if( ( (sgn(l->elem)==-sign) || (cl->weight[abs(l->elem)]==0) ) &&
            (stt->val[abs(l->elem)]==0) ) {
          pos=0;
          break;
        }

						/* add to the count of clauses and literals */
      if( pos ) {
        for(l=cl->cl[d].lett; l!=NULL; l=l->next)
          if( (sgn(l->elem)==sign) && (stt->val[abs(l->elem)]==0) )
            quanti[abs(l->elem)]++;
        positive[numpositive]=d;
        numpositive++;
      }
    }
  }

      
    				/* for each clause, take the literal with the best fraction 
				   weight/number of occurrencies */

  w=0;

  for(c=0; c<numpositive; c++) {
    d=positive[c];

    best=1000000;

    for(l=cl->cl[d].lett; l!=NULL; l=l->next)
      if( stt->val[abs(l->elem)]==0 )
        if( cl->weight[abs(l->elem)] * 1000 / quanti[abs(l->elem)] < best )
           best= cl->weight[abs(l->elem)] * 1000 / quanti[abs(l->elem)];

    w+=best;
  }

  return (w+990)/1000;
}


/*
  Lower bound based on the partition of positive clauses and the
count of occurencies of variables in positive clauses.
*/

int poscountlowerbound( setofclauses cl, clauseindex ix, state stt, int sign ) {
  int c;
  listofint l;
  int best;
  int w;

  w=0;

  for( c=firstofclass(POSCLAUSE,stt->posclause); c!=ENDOFCLASS; c=nextofclass(c,stt->posclause) ) {
    best=1000000;

    for(l=cl->cl[c].lett; l!=NULL; l=l->next)
      if( stt->val[abs(l->elem)]==0 )
        if( cl->weight[abs(l->elem)] * 1000 / stt->posocc[abs(l->elem)] < best )
           best= cl->weight[abs(l->elem)] * 1000 / stt->posocc[abs(l->elem)];

    w+=best;
  }

  return (w+990)/1000;
}


int weightlowerbound( setofclauses cl, clauseindex ix, state stt, int sign, int options ) {
  int a,b,c;

  c=poscountlowerbound( cl, ix, stt, sign );

  if( options & SELFTEST ) {
    a=allclauseslowerbound( cl, ix, stt, sign );
    b=candidatelowerbound( cl, ix, stt, sign );
    c=poscountlowerbound( cl, ix, stt, sign );
  
    if( (a!=b) || (b!=c) ) {
      printf("allclauseslowerbound: %d\n",a);
      printf("candidatelowerbound: %d\n",b);
      printf("poscountlowerbound: %d\nexiting\n",c);
      exit(1);
    }
  }

  return c;
}

/************************* End Weight Lower Bound ***************************/

