/*
  Minimal models using Davis-Putnam
Heuristics:
1. fail whenever the wieght of a model is grater or equal then the last best
2. use positive clauses to infer a lower bound of the weight
2. all the variables whose weight is more than the current difference are negated

Note: it must hold weight*1000 < overflow.

Define:

DEBUG		print debug info
INTER		print intermediate results
PRINTMODEL	print the models

NOPOSFIRST	do not use the positive heuristics while finding for the first
		model (seems to be not useful in practice, perhaps because the
		index of positive clauses must be maintained anyway)
*/


#include"list.h"
#include"queue.h"
#include"mindp.h"
#include<sys/times.h>

#define min(x,y)	(((x)<(y))?(x):(y))


state sttglobal;		/* this is the state at the last model found */

queue *q;



int usertime() {
  struct tms temp;

  times(&temp);
  return temp.tms_utime;
}


				/* print the current status of the search */

void printstatus() {
  printf("\n");
  if( !sttglobal->found ) {
    printf("No model found so far ");
    printf("(weight of models, if any, is between %d and %d).\n",sttglobal->kmin,sttglobal->kmax);
  }
  else {
    printf("Found model with weight %d ",sttglobal->kmax+1);
    printf("(weight of models is greater or equal than %d).\n",sttglobal->kmin);
  }
}


/*************************** Expand Overweigth  *****************************/
/*
  Expand by unit propagation all the variables whose
weight is greater than the current difference kmax-kmod
*/

int expandoverweight( setofclauses cl, clauseindex ix, state stt, stack stk, int kmax, int kmod, int options ) {
  int i;
  int u;

  u=0;
  i=stk->elem[stk->stackindex].firstunassweighted;

  #ifdef DEBUG
    printf("Over (%d,%d) start %d:",kmod,kmax,stk->elem[stk->stackindex].firstunassweighted);
  #endif

  while( ( cl->weight[ ix->weightorder[i] ] > kmax-kmod ) && (i<=cl->numvar) ) {
    rmallqueue(q);

    while( ( cl->weight[ ix->weightorder[i] ] > kmax-kmod ) && (i<=cl->numvar) ) {
      #ifdef DEBUG
        printf(" %d(%d)",ix->weightorder[i],cl->weight[ ix->weightorder[i] ]);
      #endif
      if( stt->val[ix->weightorder[i]] == 0 ) {
        inqueue(q,-ix->weightorder[i]);
        #ifdef DEBUG
          printf("*");
        #endif
      }
      i++;
    }
  
    if( isempty(q) )
      return 0;

    u=unitpropagation( q, cl, ix, stt, stk, options );

    if( u!=0 )
      return u;

    kmod=stk->elem[stk->stackindex].weight;
    if( kmod>kmax )
      return -1;

    #ifdef DEBUG
      printf("\nOver again (%d,%d) start %d:",kmod,kmax,i);
    #endif
  }

  stk->elem[stk->stackindex].firstunassweighted=i;

  #ifdef DEBUG
    printf("end %d\n",stk->elem[stk->stackindex].firstunassweighted);
  #endif
  
  return u;
}

/************************* End Expand Overweigth  ***************************/


/********************************* Weight ***********************************/
/*
  Weightval: weight of the positive atoms of a model (not used any more).
*/

int weightval( setofclauses cl, state stt ) {
  int i;
  int w;

  w=0;

  for(i=1; i<=cl->numvar; i++)
    if( stt->val[i]==1 )
      w+=cl->weight[i];

  return w;
}

/******************************* End Weight *********************************/


/****************************** First & Next ********************************/
/*
  First: find the first model with weight less or equal than kmax
  Next: find the next model with weight less or equal then kmax

The first functions builds the clauseindex, the state, the stack,
and the structures needed for handling weights (that is, the weight
order in ix, the positive count in stt, etc.) propagate the unit
clauses (if the PROPAGATE options is active), simplify the set
of clauses (if the SIMPLIFY options is active), and then uses
the next function for returning the first model.

The next function need not to be called with the same value of
kmax. The only constraint is that the value of kmax must be
less or equal than the value used for determining stt and stk.

NOTE: next do not make backtracking: as a result, if it is used
for restarting a search, it must be that either 1. the current
state is inconsistent with kmax, or 2. the calling function
performs a call to backtracking prior to calling next.
*/

/*
  Find the first model with weight less or equal than kmax
*/

int find( setofclauses cl, clauseindex ix, state stt, stack stk, int kmax, int options ) {
  int u;
  int lit;
  int kmod,klower;

  while(1) {
    #ifdef DEBUG
      printf("Model before dp: ");
      printstate( stt );
      fflush(NULL);
    #endif

    if( options & SELFTEST )
      selftest( cl, ix, stt, stk, options );

    rmallqueue(q);
    inqueue(q,stk->elem[ stk->stackindex ].literal);
    u=unitpropagation(q, cl, ix, stt, stk, options);
    kmod=stk->elem[stk->stackindex].weight;

    if( options & SELFTEST ) {
      selftest( cl, ix, stt, stk, options );
      if( u!=-1 )
        consselftest( cl, ix, stt, stk, options );
    }

    #ifdef DEBUG
      printf("UP EXECUTED\n");
      printf("Up=%d Kmod=%d\n",u,kmod);
      printf("Model after up: ");
      printstate( stt );
      fflush(NULL);
    #endif

    if( kmod > kmax )
      u=-1;

    klower=0;
    if( options & POSITIVE ) {
      if( u>-1 ) {
        klower=weightlowerbound( cl, ix, stt, 1, options );
        if( kmod+klower > kmax )
          u=-1;
      }
    }

    if( options & OVER ) {
      if( u==0 ) {
        u=expandoverweight( cl, ix, stt, stk, kmax, kmod, options );
        kmod=stk->elem[stk->stackindex].weight;
        if( kmod > kmax )
          u=-1;
      }
    }

    #ifdef DEBUG
      printf("State after heuristics\n");
      printf("Up=%d Weight=%d Estimated=Weight+%d\n",u,kmod,klower);
      printf("Model after heuristics: ");
      printstate( stt );
      printstack( stk );
      fflush(NULL);
    #endif

    if( u==1 )
      return kmod;
    else if( u==-1 ) {
      if( backtracking( cl, ix, stt, stk, options ) == -1 )
        return -1;
      }
    else {
      stt->branch++;
      if( options & SELFTEST )
        consselftest( cl, ix, stt, stk, options );
      lit=choice( cl, ix, stt, stk, options );
      addlevel( stk, lit, 0, options );
    }
  }

}


/*
  First model whose weight is less or equal than kmax.
*/

int first( setofclauses cl, clauseindex *ix, state *stt, stack *stk, int kmax, int options ) {
  int k,kmin;
  int x;
  int u;


				/* initialization of structures */

  #ifdef DEBUG
    printf("Initializations\n");
  #endif

  initial( cl, ix, stt, stk, options );

  (*stt)->weighted=1;


				/* extra structures needed for weight */

  #ifdef DEBUG
    printf("Initialization weight structures\n");
  #endif

  makeorder( cl, *ix );
  if( options & POSITIVE ) {
    makecand( cl, *ix );
    makeposlits( cl, *stt );
    makeposclause( cl, *stt );
  }
  if( options & SELFTEST )
    selftest( cl, *ix, *stt, *stk, options );


				/* simplify literals */

  if( options & SIMPLIFY )
    if( simplify( cl, *ix, *stt, *stk, options )==1 )
      return (*stk)->elem[(*stk)->stackindex].weight;

				/* propagate unit clauses */

  if( options & PROPAGATE ) {
    #ifdef DEBUG
      printf("Propagating unit clauses\n");
    #endif
    u=propagateunitclauses( cl, *ix, *stt, *stk, options );

    if( u==-1 )
      return -1;
    else if( u==1 )
      return (*stk)->elem[(*stk)->stackindex].weight;
  }


				/* create first level of the stack */

  #ifdef DEBUG
    printf("Creating first level of the stack\n");
  #endif
  firstlevel( cl, *ix, *stt, *stk, options );



				/* upper and lower bounds */

  #ifdef DEBUG
    printf("Initial estimates of weight\n");
  #endif

  k=0;
  for(x=1; x<=cl->numvar; x++)
    k+=cl->weight[x];
  kmin=0;

  if( options & POSITIVE ) {
    k=k-allclauseslowerbound( cl, *ix, *stt, -1 );
    kmin=kmin+weightlowerbound( cl, *ix, *stt, 1, options );
  }

  kmax=min(kmax,k);

  (*stt)->kmax=kmax;
  (*stt)->kmin=kmin;


				/* first model */

  #ifdef NOPOSFIRST
    return find( cl, *ix, *stt, *stk, kmax, options&~POSITIVE );
  #else
    return find( cl, *ix, *stt, *stk, kmax, options );
  #endif
}


/*
  Next model with weight less or equal than kmax
*/

int next( setofclauses cl, clauseindex ix, state stt, stack stk, int kmax, int options ) {
  #ifdef DEBUG
    printf("\nStarting new search.\n\n");
  #endif

  if( backtracking( cl, ix, stt, stk, options ) == -1 )
    return -1;

  return find( cl, ix, stt, stk, kmax, options );
}

/**************************** End First & Next ******************************/


/******************************** Bounded ***********************************/
/*
  Functions for determining models with weight less or equal than kmax.

bounded: check existence and find the first only

firstbounded: check existence and find the first model

nextbounded: next model with weight less or equal than kmax
*/


/*
  Bounded: check existence and find the first only
*/

int bounded( setofclauses cl, state *stt, int kmax, int options ) {
  clauseindex ix;
  stack stk;
  int u;

  u=first( cl, &ix, stt, &stk, kmax, options );

  freeindex(ix);
  freestack(stk);

  return u;
}


/*
  First model bounded by kmax
*/

int firstbounded( setofclauses cl, clauseindex *ix, state *stt, stack *stk, int kmax, int options ) {
  return first( cl, ix, stt, stk, kmax, options & ~SIMPLIFY );
}


/*
  Next model bounded
*/

int nextbounded( setofclauses cl, clauseindex ix, state stt, stack stk, int kmax, int options ) {
  return next( cl, ix, stt, stk, kmax, options & ~SIMPLIFY );
}

/****************************** End Bounded *********************************/


/******************************** Minimal ***********************************/
/*
  Functions for determining minimal models.

findfirstminweight: this is the function that actually determine the minimal
weight and the first minimal model. This cannot be left for external use,
since SIMPLIFY must be reset if looking for all minimal models.

minweight: allocates index and stack, and return the first minimal model
(it is not possible to find other minimal models this way).

firstminweight: this is a simple call to the above functions. SIMPLIFY is
reset before the call.

nextminweight: backtracking + call to next (SIMPLIFY is reset)
*/

/*
  Find the first minimal model
*/

int findfirstminweight( setofclauses cl, clauseindex *ix, state *sttlast, stack *stklast, int options ) {
  state stt;		/* state at the last model found */
  stack stk;		/* stack at the last model found */
  int k,kmax,kmin,klast;
  int x;
  int u;
  int inittime;


				/* allocate memory for the last state and stack */

  *sttlast=(state) malloc( sizeof(**sttlast) );
  if( *sttlast==NULL ) {
    fprintf(stderr,"Malloc failed in mindp() while allocating sttlast\n");
    exit(1);
  }
  *stklast=(stack) malloc( sizeof(**stklast) );
  if( *stklast==NULL ) {
    fprintf(stderr,"Malloc failed in mindp() while allocating stklast\n");
    exit(1);
  }


				/* initial weight (first() will check if something better
				   can be done) */

  kmax=0;
  for(x=1; x<=cl->numvar; x++)
    kmax+=cl->weight[x];
  

				/* find first model (if any) */

  #ifdef DEBUG
    printf("\n***********************\n");
    printf("* FINDING FIRST MODEL *\n");
    printf("***********************\n\n");
  #endif

  sttglobal=*sttlast;
  sttglobal->found=0;
  (*sttlast)->intertime=NULL;
  (*sttlast)->intertime=conslistofint((*sttlast)->intertime,usertime());
  (*sttlast)->intermod=NULL;

  k=first( cl, ix, &stt, &stk, kmax, options );

  stt->found=1;

  stt->intertime=conslistofint((*sttlast)->intertime,usertime());
  stt->intermod=conslistofint((*sttlast)->intermod,k);

  copystate( *sttlast, stt );
  copystack( *stklast, stk );

  if( k<0 )
    return k;

  klast=k;

  #ifdef INTER
    printf("K=%d  TIME=%d\n",k, (*sttlast)->intertime->elem );
    /* printweighted( cl, stt ); */
  #endif


				/* main loop */

  while( klast>stt->kmin ) {
    #ifdef DEBUG
      printf("\n**********************\n");
      printf("* FINDING NEXT MODEL *\n");
      printf("**********************\n\n");
    #endif

    kmax=k-1;
    stt->kmax=k-1;

                                /* next model */
    k=next( cl, *ix, stt, stk, kmax, options );
    if( k==-1 )
      break;

    klast=k;
    stt->intertime=conslistofint((*sttlast)->intertime,usertime());
    stt->intermod=conslistofint((*sttlast)->intermod,k);
 
    copystate( *sttlast, stt );
    copystack( *stklast, stk );

    #ifdef INTER
      printf("K=%d  TIME=%d\n",k, (*sttlast)->intertime->elem );
      /* printweighted( cl, stt ); */
    #endif
  }

  return klast;
}




/*
  Only the first model is necessary
*/

int minweight( setofclauses cl, state *sttlast, int options ) {
  clauseindex ix;
  stack stklast;
  int u;

  u=findfirstminweight( cl, &ix, sttlast, &stklast, options );

  freeindex(ix);
  freestack(stklast);

  return u;
}


/*
  Find first minimal model
*/

int firstminweight( setofclauses cl, clauseindex *ix, state *sttlast, stack *stklast, int options ) {
  return findfirstminweight( cl, ix, sttlast, stklast, options&~SIMPLIFY );
}


/*
  Next model with minimal weight
*/

int nextminweight( int kmax, setofclauses cl, clauseindex ix, state stt, stack stk, int options ) {
  return next( cl, ix, stt, stk, kmax, options&~SIMPLIFY );
}

/****************************** End Minimal *********************************/



