/*
  Minimal models using Davis-Putnam
Heuristics:
1. fail whenever the wieght of a model is grater or equal then the last best
2. use positive clauses to infer a lower bound of the weight
2. all the variables whose weight is more than the current difference are negated

Note: it must hold weight*1000 < overflow.

Define:

SELFTEST	always check dynamically created structures

OVER		all variables whose weight is greater then the difference between
		the current weight and the maximum are set to false
POSITIVE	use heuristics on positive clauses

PROPAGATE	propagate the value of unit clauses before starting the search
		(useful if TRIVIALCHOICE or WEIGHTCHOICE are used)

DEBUG		print debug info
INTER		print intermediate results
PRINTMODEL	print the models
*/


#include"list.h"
#include"queue.h"
#include"mindp.h"


/****************************** Init Weight *********************************/
/*
  Creates the index of the candidate clauses. These are the clauses
that contain at least a pair of positive literals with weight > 0
*/

int makecand( setofclauses cl, clauseindex ix ) {
  int c,h;
  listofint l;

  ix->numcand=0;

  for(c=0; c<cl->numclauses; c++) {
    h=0;
    for(l=cl->cl[c].lett; l!=NULL; l=l->next)
      if( (l->elem > 0) && (cl->weight[abs(l->elem)]>0) )
        h++;

    if( h>=2 ) {
      ix->cand[ix->numcand]=c;
      ix->numcand++;
    }
  }
  return 0;
}


/*
  Create the index of the variables (ordered by weight).
*/

int makeorder( setofclauses cl, clauseindex ix ) {
  int i,j,k;

  for(i=1; i<=cl->numvar; i++)
    ix->weightorder[i]=i;

  for(i=1; i<cl->numvar; i++)
    for(j=i+1; j<=cl->numvar; j++)
      if( cl->weight[ ix->weightorder[i] ] <  cl->weight[ ix->weightorder[j] ] ) {
        k=ix->weightorder[i];
        ix->weightorder[i]=ix->weightorder[j];
        ix->weightorder[j]=k;
      }
   
  return 0;
}


/*
  Create the array of the number of positive atoms in each clause.
*/

int makeposlits( setofclauses cl, state stt ) {
  int c;
  listofint l;
  int np;

  for(c=0; c<cl->numclauses; c++) {
    np=0;

    for(l=cl->cl[c].lett; l!=NULL; l=l->next)
      if( (l->elem>0) && (cl->weight[abs(l->elem)]>0) )
        np++;

    stt->poslits[c]=np;

    #ifdef DEBUG
      printf("numposlits(%d)=%4d\t",c,np);
    #endif
  }

  #ifdef DEBUG
    printf("\n");
  #endif

  return 0;
}

/*
  Create the initial partition positive/nonpositive, and
the array of the number of occurences of literals in positive clauses.
*/

int makeposclause( setofclauses cl, state stt ) {
  int c,i;
  listofint l;

  stt->posclause=newpartition();
  initpartition(cl->numclauses, 2, stt->posclause);

  memset( stt->posocc, 0, (cl->numvar+1)*sizeof(int) );

  for(c=0; c<cl->numclauses; c++)
    if( cl->cl[c].numlett == stt->poslits[c] ) {
      moveelem(c, POSCLAUSE, stt->posclause );

      for( l=cl->cl[c].lett; l!=NULL; l=l->next )
        stt->posocc[l->elem]++;
    }

  #ifdef DEBUG
    for(i=1; i<=cl->numvar; i++)
      printf("posocc[%d]=%d\t",i,stt->posocc[i]);
    printf("\n");
  #endif

  return 0;
}

/**************************** End Init Weight *******************************/


