/*
  Minimal models using Davis-Putnam
Heuristics:
1. fail whenever the weight of a model is greater or equal than the last best found
2. use positive clauses to infer a lower bound of the weight
2. all the variables whose weight is more than the current difference are set to false

Note: it must hold weight*1000 < overflow.

Define:

USESIGNALS	compile with the functions for handling signals

*/

#include"dp.h"

#include<signal.h>

#define PIDFILENAME "pid.mindp"


                                /* signal handlers */

void onexit() {
  unlink(PIDFILENAME);
}

#ifdef USESIGNALS

  void usersignal( int s ) {
    printstatus();

    signal( SIGUSR1, usersignal);
  }

  void exitsignal( int s ) {
    printf("\nSignal %d caught. Exiting.\n",s);
    printstatus(s);
    onexit();

    exit(1);
  }
#endif

/* 
  check whether the model satisfies the set of clauses
*/

int verifymodel( setofclauses cl, state stt ) {
  int c;
  int x;
  listofint l;
  int sat,this;

  sat=1;

  for(c=0; c<cl->numclauses; c++) {
    this=0;

    for(l=cl->cl[c].lett; l!=NULL; l=l->next)
      if( l->elem>0 ) {
        if( valueofvar(l->elem,stt)==1 )
          this=1;
      }
      else
        if( valueofvar(-l->elem,stt)==-1 )
          this=1;

    if( this==0 )
      sat=0;
  }

  return sat;
}


/*
  Verify the weight of the model
*/

int verifyweight( setofclauses cl, state stt ) {
  int w;
  int x;

  w=0;

  for(x=1; x<=cl->numvar; x++)
    if( valueofvar(x,stt)==1 )
      w+=cl->weight[x];

  return w;
}


main( int argc, char *argv[] ) {
  int c;
  setofclauses cl;
  clauseindex ix;
  state stt;
  stack stk;
  int k;
  listofint l,r;

  FILE *pidfile;
  int starttime, endtime;


  cl=newsetofclauses();
  readclauses( stdin, &cl );

  /* printclauses( cl ); */

  #ifdef DEBUG
    printclauses( cl );
  #endif

  #ifdef USESIGNALS
    pidfile=fopen(PIDFILENAME,"w");
    fprintf(pidfile,"%d",getpid());
    fclose(pidfile);
  
    signal( SIGTERM, exitsignal );
    signal( SIGINT, exitsignal );
    signal( SIGUSR1, usersignal );
  #endif

  if( argc>1 ) {
    if( onemodel(cl,&stt,DEFAULTDP)==1 )
      printf("SAT\n");
    else
      printf("UNSAT\n");

    printf("%d\n",stt->branch);

    return 0;
  }

  starttime=usertime();
  k=minweight( cl, &stt, DEFAULTDP );
  endtime=usertime();

  if( k<0 )
    printf("Inconsistente\n");
  else
    printf("Consistente: peso=%d\n",k);

//  fprintf(stderr,"%d\n",stt->branch);

//  fprintf(stderr,"nodes: %d\n",stt->branch);


  fprintf(stderr,"%4d ",endtime-starttime);

#ifdef NO
  /* r=reverselistofint(stt->intertime); */
  if( stt->intertime->next==NULL ) 
    fprintf(stderr,"%4d\n",endtime-starttime);
  else {
    fprintf(stderr,"%4d ",endtime - stt->intertime->elem );
    for(l=stt->intertime; l->next!=NULL; l=l->next)
      fprintf(stderr,"%4d ",l->elem - l->next->elem );
    fprintf(stderr,"\n");
  }

  fprintf(stderr,"%4d %4d ",k,k);
  for(l=stt->intermod; l!=NULL; l=l->next)
    fprintf(stderr,"%4d ",l->elem);
#endif

  if( k>=0 ) {
  if( verifymodel( cl, stt )==0 )
    fprintf(stderr,"error");
  else if( k!=verifyweight(cl,stt) )
    fprintf(stderr,"error weight");
  }

  fprintf(stderr,"\n");

  onexit();

  return 0;
}

