/*
  Unit propagation and stack handling.
*/

#include"list.h"
#include"queue.h"
#include"mindp.h"


/************************* Update Binary Count ****************************/
/*
 Modify the binary count of a clause. This should be done only when the
clause reach exactly the count of two undetermined literals (all other
being false), or immediately after it becomes no more a binary clause.
*/

int updatebinarycount( int c, setofclauses cl, state stt ) {
  listofint l;
  int u;

  if( (stt->truelits[c]==0) && (stt->falselits[c]==cl->cl[c].numlett-2 ) )
    u=1;
  else
    u=-1;

  for(l=cl->cl[c].lett; l!=NULL; l=l->next) {
    if( stt->val[abs(l->elem)]==0 )
      if( l->elem > 0 )
        stt->binarypos[l->elem]+=u;
      else
        stt->binaryneg[-l->elem]+=u;
  }

  stt->binary+=u;

  return u;
}

/*********************** End Update Binary Count **************************/


/************************ Update Positive Count ***************************/
/*
  Modify the positive count of a clause. This should be done whenever a
literal of the clause changes its signs (any literal, including those
with weight =0). This should be called only when a clause changes from
positive to negativa and viceversa, as it modifies the count of positive
occurrences of variables.
*/

/*
  We use stt->posclause==NULL instead of options&POSITIVE because this allows
maintaining the partition of positive clauses even if the POSITIVE heuristics
is not used. For instance, when finding the first model the heuristics is not
useful, but the partition has to be maintained if we want to use the heuristics
for finding further models.
*/

int updateposcount( int c, setofclauses cl, state stt ) {
  listofint l;
  int p;

		/* if posclause=NULL, the partion of positive clauses is not used
		   (for instance, the POSITIVE bit is 0) */

  if( ( stt->weighted==0 ) || (stt->posclause==NULL) )
    return 0;

  if( ( stt->truelits[c] == 0 ) && ( cl->cl[c].numlett - stt->falselits[c] == stt->poslits[c] ) )
    p=POSCLAUSE;
  else
    p=NONPOSCLAUSE;

  if( (p==POSCLAUSE) && (whichclass(c,stt->posclause)==NONPOSCLAUSE) ) {
    moveelem( c, POSCLAUSE, stt->posclause );

    for(l=cl->cl[c].lett; l!=NULL; l=l->next)
      if( stt->val[abs(l->elem)]==0 )
        stt->posocc[abs(l->elem)]++;
  }

  if( (p==NONPOSCLAUSE) && (whichclass(c,stt->posclause)==POSCLAUSE) ) {
    moveelem( c, NONPOSCLAUSE, stt->posclause );

    for(l=cl->cl[c].lett; l!=NULL; l=l->next)
      if( (l->elem>0) && (cl->weight[abs(l->elem)]>0) && (stt->val[abs(l->elem)]==0) )
        stt->posocc[abs(l->elem)]--;
  }

  return 0;
}

/********************** End Update Positive Count *************************/


/*************************** Unit Propagation *****************************/
/*
  Unit propagation procedure.
Iteratively extract a literal from the queue (until it is empty).
Set the value of the literal, and check the set of clauses in which
the variable is present.

Clauses in which the literal has the same sign: increase the truelits
count of the clause. If it is equal to the size of the clause, then
the number of satisfied clauses is increased.

Clauses in which the literal has opposite sign: increase the falselits
of the clause; if this number is equal to the number of literals, the
clause is false; if all the literals but one are false (and the other
is not true), then insert the remaining literal in the queue.

The monotone literal rule is part of this procedure, and it is used
only if options&SIMPLIFY != 0. This is done here only for the literals
of clauses that becomes true. A separate procedure would require a
linear check of stt->pos[i] and stt->neg[i] for any unassigned
variable.
*/

int unitpropagation( queue *q, setofclauses cl, clauseindex ix, state stt, stack stk, int options ) {
  int v,a,c,u;
  listofint lp,ln,ll,lc,l;
  int contradiction;
  int stkidx;

  #ifdef UDEBUG
    printf("[");
    for(v=1; v<=cl->numvar; v++)
      printf("%d=%d ",v,stt->val[v]);
    printf("]\n");

    printf("queue: ");
    printqueue(q);
  #endif


  contradiction=0;

  stkidx=stk->stackindex;

  while( !isempty(q) ) {
				/* new literal from the queue */
    v=outqueue(q);
    a=abs(v);

    #ifdef UDEBUG
      printf("top=%d, queue: ",v);
      printqueue(q);
    #endif

				/* if it is already set, check the sign */
    if( stt->val[a] ) {
      if( stt->val[a] != sgn(v) )
        return -1;
    }
    else {
				/* the literal is deduced */
      stk->deduced[stk->elem[stkidx].firstdeduced+stk->elem[stkidx].numdeduced]=v;
      stk->elem[stkidx].numdeduced++;

      if( v>0 )
        stk->elem[stkidx].weight+=cl->weight[v];

				/* set the value of the variable */
      stt->val[a]=sgn(v);

				/* list of clauses in which v is negative or positive */
      if( pos(v) ) {
        lp=ix->positive[a];
        ln=ix->negative[a];
      }
      else {
        lp=ix->negative[a];
        ln=ix->positive[a];
      }

				/* positive: the true count is increased */

      for(ll=lp; ll!=NULL; ll=ll->next) {
        stt->truelits[ll->elem]++;

        if( stt->truelits[ll->elem] == 1 ) {		/* previously clause was not true */
          #ifdef UDEBUG
            printf("Clause %d is now true: ",ll->elem);
            printclause( cl, ll->elem );
          #endif


          stk->elem[stkidx].numtrueclauses++;

          if( options & SIMPLIFY )
            for(l=cl->cl[ll->elem].lett; l!=NULL; l=l->next)
              if( stt->val[abs(l->elem)]==0 ) 
                if( l->elem > 0 ) {
                  stt->pos[l->elem]--;
                  if( (stt->pos[l->elem]==0) && (stt->neg[l->elem]!=0) ) {
                    inqueue(q,-l->elem);
                    #ifdef UDEBUG
                      printf("Literal %d deduced \n",topqueue(q));
                    #endif
                  }
                }
                else {
                  stt->neg[-l->elem]--;
                  if( stt->neg[-l->elem]==0 && stt->pos[-l->elem]!=0 &&
                      ( cl->weight[-l->elem]==0 || stt->weighted==0 ) ) {
                    inqueue(q,-l->elem);
                    #ifdef UDEBUG
                      printf("Literal %d deduced \n",topqueue(q));
                    #endif
                  }
                }


          if( stt->falselits[ll->elem] == cl->cl[ll->elem].numlett - 2 )
            updatebinarycount( ll->elem, cl, stt );
    
          updateposcount( ll->elem, cl, stt );
        }
      }


				/* negative: the false count is increased */

      for(ll=ln; ll!=NULL; ll=ll->next) {
        c=ll->elem;

        stt->falselits[c]++;
  
						/* all false: contradiction */
        if( stt->falselits[c]==cl->cl[c].numlett ) {
          #ifdef UDEBUG
            printf("Clause %d is false\n",c);
          #endif
          contradiction=1;
        }

						/* one unassigned, all false: new literal */
        if( (stt->falselits[c]==cl->cl[c].numlett-1) && (stt->truelits[c]==0) )
          for(lc=cl->cl[c].lett; lc!=NULL; lc=lc->next)
            if( stt->val[abs(lc->elem)]==0 ) {
              inqueue(q,lc->elem);
              #ifdef UDEBUG
                printf("deduced literal %d from clause ",lc->elem);
                printclause( cl, c );
              #endif
              break;
            }

						/* two unassigned literals: update binary count */
        if( (stt->falselits[c]==cl->cl[c].numlett-2) && (stt->truelits[c]==0) )
          updatebinarycount( c, cl, stt );

						/* update positive count */
        if( (v<0) && (cl->weight[abs(v)]>0) )
          stt->poslits[ll->elem]--;

						/* update positive count */
        updateposcount(ll->elem, cl, stt);
      }

    }  /* end else not set */

    #ifdef UDEBUG
      printf("--------------------------\n");
      selftest(cl,ix,stt,stk);
    #endif

    if( contradiction )
      return -1;

  }  /* end while */

  if( stk->elem[stkidx].numtrueclauses == cl->numclauses )
    return 1;
  else
    return 0;
}

/************************* End Unit Propagation ***************************/


/****************************** Add Level *********************************/
/*
  Expand a new stack level, using the literal l as a new choice.
*/

int addlevel( stack stk, int lit, int tried, int options ) {
  int stkind;

  stk->stackindex++;

  stkind=stk->stackindex;

  stk->elem[stkind].literal=lit;
  stk->elem[stkind].triedneg=tried;

  stk->elem[stkind].firstdeduced=stk->elem[stkind-1].firstdeduced+stk->elem[stkind-1].numdeduced;
  stk->elem[stkind].numdeduced=0;

  stk->elem[stkind].weight=stk->elem[stkind-1].weight;

  stk->elem[stkind].firstunassweighted=stk->elem[stkind-1].firstunassweighted;

  stk->elem[stkind].numtrueclauses=stk->elem[stkind-1].numtrueclauses;
  
  return 0;
}
  
/**************************** End Add Level *******************************/


/***************************** Remove Level *******************************/
/*
  Remove the last inserted level of the stack.
*/

int rmlevel( setofclauses cl, clauseindex ix, state stt, stack stk, int options ) {
  int stkind;
  int x,v;
  listofint lp,ln,l,lc,ll;
  int lit;

  stkind=stk->stackindex;


			/* per ognuno dei letterali dedotti, fa le modifiche alla rovescia */


  for(x=stk->elem[stkind].firstdeduced+stk->elem[stkind].numdeduced-1;
      x>=stk->elem[stkind].firstdeduced;
      x--) {

    v=stk->deduced[x];

			/* sottrae gli opportuni valori ai clause count */

    if( v>0 ) {
      lp=ix->positive[abs(v)];
      ln=ix->negative[abs(v)];
    }
    else {
      lp=ix->negative[abs(v)];
      ln=ix->positive[abs(v)];
    }

    for(l=ln; l!=NULL; l=l->next) {
      stt->falselits[l->elem]--;

      if( (stt->truelits[l->elem]==0) && (stt->falselits[l->elem]==cl->cl[l->elem].numlett-3) )
        updatebinarycount( l->elem, cl, stt );

      if( (v<0) && (cl->weight[abs(v)]>0) )
        stt->poslits[l->elem]++;

      updateposcount( l->elem, cl, stt );
    }

    for(l=lp; l!=NULL; l=l->next) {
      stt->truelits[l->elem]--;

      if( (stt->truelits[l->elem]==0) && (stt->falselits[l->elem]==cl->cl[l->elem].numlett-2) )
        updatebinarycount( l->elem, cl, stt );

      if( options & SIMPLIFY )
        if( stt->truelits[l->elem]==0 )
          for(ll=cl->cl[l->elem].lett; ll!=NULL; ll=ll->next)
            if( stt->val[abs(ll->elem)]==0 )
              if( ll->elem > 0 )
                stt->pos[ll->elem]++;
              else
                stt->neg[-ll->elem]++;

      updateposcount( l->elem, cl, stt );
    }


                        /* rimette a 0 i letterali che sono stati determinati a questo passo */
			/* va alla fine perche' le modifiche vanno fatte alla rovescia */
    stt->val[ abs(v) ]=0;

  }

  stk->stackindex--;

  return stk->stackindex;
}

/*************************** End Remove Level *****************************/


/***************************** Backtracking *******************************/
/*
  Backtracking of the stack: run rmlevel until either a literal is found
whose negation has not been tried yet, or the stack is over (ie all the
search tree has been searched).
*/

int backtracking( setofclauses cl, clauseindex ix, state stt, stack stk, int options ) {
  int stkind;
  int x,v;
  listofint lp,ln,l;
  int lit;

  stkind=stk->stackindex;

  do {

    stkind=rmlevel( cl, ix, stt, stk, options );

    if( options & SELFTEST )
      if( stkind>0 )
        selftest(cl,ix,stt,stk,options);

  } while( ( stkind>=0 ) && ( stk->elem[stkind+1].triedneg==1 ) );

  if( stkind<0 )
    return -1;
  
  stk->stackindex=stkind;
  lit=-stk->elem[stkind+1].literal;		/* if stack is implemented as a list, this is an error */

  addlevel( stk, lit, 1, options );

  return 0;
}

/*************************** End Backtracking *****************************/

